# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 hankei6km
# Licensed under the MIT License. See LICENSE.txt in the project root.

import os
import sys
import shlex
import re
import json
import errno
import inspect
from operator import itemgetter
from functools import wraps

from fuse import FUSE, FuseOSError, Operations, LoggingMixIn
from pathlib import Path


class XrosFS(LoggingMixIn, Operations):
    def __init__(self, shell, cmd_builder, root):
        self.shell = shell
        self.cmd_builder = cmd_builder
        self.root = os.path.normpath(root)

    opts2statvfs_flag = {
        'append': os.ST_APPEND,
        'nodev': os.ST_NODEV,
        'nosuid': os.ST_NOSUID,
        'synchronous': os.ST_SYNCHRONOUS,
        'mandlock': os.ST_MANDLOCK,
        'nodiratime': os.ST_NODIRATIME,
        'rdonly': os.ST_RDONLY,
        'write': os.ST_WRITE,
        'noatime': os.ST_NOATIME,
        'noexec': os.ST_NOEXEC,
        'relatime': os.ST_RELATIME,
        'rw': 32  # os.ST_???
    }

    errno_mapper = {
        '': [(
            re.compile('Permission denied$'), 1
        ), (
            re.compile(': File exists$'), 17
        ), (
            re.compile(': Directory not empty$'), 39
        ), (
            re.compile('No such file or directory$'), 2
        )],
        'fsync': [(
            re.compile(': File exists$'), 17
        ), (
            re.compile(': Directory not empty$'), 39
        ), (
            re.compile('No such file or directory$'), 2
        )]
    }

    def _is_directory_traversal(self, path, full_path):
        """ Detect directory traversal.
        """
        norm_path = os.path.normpath(full_path)
        root = os.path.commonpath([self.root, norm_path])
        return root != self.root

    def _full_path(self, path):
        if os.path.isabs(path):
            p = Path(path)
            if p.root != path:  # confirmed that path is abs.
                path = path[len(p.root):]
            else:
                path = ''
        return os.path.join(self.root, path)

    def full_path(p=['path']):
        def full_path_inner(f):
            # https://stackoverflow.com/questions/218616/getting-method-parameter-names-in-python
            path_args_list = []
            spec_args = inspect.getargspec(f).args
            for i in p:
                # Build prefix of full_path arg name.
                # `path` -> `full_path`
                # `foo` -> `full_path_foo'
                # `bar` -> `full_path_bar'
                full_path_arg_prefix = \
                    'full_' if i == 'path' else 'full_path_'
                # Pickup target args that passed name of args stored path.
                # It is support only positional argument yet.
                path_args_list.append((
                    i,
                    full_path_arg_prefix + i,
                    spec_args.index(i)))

            @wraps(f)
            def ret(self, *args, **kwargs):
                # Add full_path into kwargs.
                for arg_name, arg_full_path_name, arg_idx in path_args_list:
                    # Shift position of idx by considering `self` arg.
                    path = args[arg_idx - 1]
                    full_path = self._full_path(path)
                    if self._is_directory_traversal(path, full_path):
                        raise PermissionError(
                            'Directory traversal deteceted, '
                            'arg ' + arg_name + '=' + path)
                    kwargs[arg_full_path_name] = full_path
                return f(self, *args, **kwargs)

            return ret
        return full_path_inner

    def _raise_unknown(self, command, res):
        raise RuntimeError('cmd:{0} errno:{1} stderr:{2}'.format(
            command[0], res.errno, res.stderr))

    def _entry(self,
               command, errno_map='', raise_unknow=True, *args, **kwargs):
        res = self.shell.entry(command, *args, **kwargs)
        if res.errno != 0:
            for r, fuse_errno in self.errno_mapper[errno_map]:
                if r.search(res.stderr, 0) is not None:
                    raise FuseOSError(fuse_errno)
            if raise_unknow:
                self._raise_unknown(command, res)

        return res

    def _get_homedir(self):
        res = self._entry(self.cmd_builder._get_homedir(), quote=False)
        return res.stdout.split('\n', 1)[0]

    @full_path()
    def access(self, path, mode, full_path=None):
        # Map mode bit to each `test` command
        mode_test = [{
            'mode': os.X_OK,
            'test': ['-x', full_path]
        }, {
            'mode': os.W_OK,
            'test': ['-w', full_path]
        }, {
            'mode': os.R_OK,
            'test': ['-r', full_path]
        }]
        # Append `test` command to list by mode
        test_commands = []
        if mode == os.F_OK:
            test_commands = [self.cmd_builder.access('-e', full_path)]
        else:
            for m in mode_test:
                if mode & m['mode'] == m['mode']:
                    test_commands.append(self.cmd_builder.access(*m['test']))

        # Build command list from `test` commands list
        command = []
        for c in test_commands:
            for i in self.shell.quote(c):
                command.append(i)
            command.append('&&')
        if len(command) > 0:
            res = self._entry(command[:-1], quote=False, raise_unknow=False)
            if res.errno != 0:
                raise FuseOSError(errno.EACCES)
        # else:
        #     raise RuntimeError('Can\'t parse mode:' + mode)

    @full_path()
    def chmod(self, path, mode, full_path=None):
        res = self._entry(self.cmd_builder.chmod(mode, full_path))
        return res.errno

    @full_path()
    def chown(self, path, uid, gid, full_path=None):
        res = self._entry(self.cmd_builder.chown(uid, gid, full_path))
        return res.errno

    @full_path()
    def getattr(self, path, fh=None, full_path=None):
        command = self.cmd_builder.getattr(full_path)
        res = self._entry(
            command, raise_unknow=False
        )
        # TODO: remove bellow code that recover from getatter failed.
        # what reason to occurs stdout and stderr is balnk with errno ==1.
        # why??
        cnt = 1
        while res.stdout == '' and res.errno == 1 and res.stderr == '':
            # print('*************** retry getattr')
            res = self._entry(
                command, raise_unknow=False
            )
            cnt = cnt + 1
            if cnt > 9:
                raise RuntimeError('Fatale error getattr failed 10 times')
        if res.errno != 0:
            self._raise_unknown(command, res)

        try:
            st = json.loads(res.stdout)
            # hexadecimal string to integer.
            # (json not support hexadecimal literal).
            st['st_mode'] = int(st['st_mode'], 16)
        except json.JSONDecodeError as e:
            raise RuntimeError('stat: ' + full_path) from e

        return st

    @full_path()
    def readdir(self, path, fh, full_path=None):
        # TODO: Support parse filename that contains '\n'.
        # And test another control chars (i.e. '\0' etc.).
        res = self._entry(self.cmd_builder.readdir(full_path))
        dirents = re.sub(r'\n$', '', res.stdout).split('\n')
        for r in dirents:
            yield r

    @full_path()
    def readlink(self, path, full_path):
        res = self._entry(self.cmd_builder.readlink(full_path))
        ret = res.stdout
        # TODO: Enable bellow snitaize code and check direcotry traversal.
        # if os.path.isabs(ret):
        #     # Path name is absolute, sanitize it.
        #     ret = os.path.relpath(ret, self.root)
        return ret

    @full_path()
    def rmdir(self, path, full_path):
        res = self._entry(self.cmd_builder.rmdir(full_path))
        return res.errno

    @full_path()
    def mkdir(self, path, mode, full_path):
        res = self._entry(self.cmd_builder.mkdir(full_path))
        errno = res.errno
        if res.errno == 0:
            errno = self.chmod(path, mode)
        return errno

    @full_path()
    def statfs(self, path, full_path):
        res = self._entry(self.cmd_builder.statfs(full_path))
        if res.errno != 0:
            # not found ??
            raise FuseOSError(errno.ENOENT)
        # try:
        #     st = json.loads(res.stdout)
        # except json.JSONDecodeError:
        #     raise FuseOSError(errno.ENOENT)
        st = json.loads(res.stdout)

        # retrive mount flag
        # (form remote is right?)
        mounts = self._entry(['cat', '/proc/mounts']).stdout
        li = sorted([
            itemgetter(1, 3)(i.split(' '))
            for i in mounts.split('\n')
            if len(i) > 0 and full_path.startswith(i.split(' ', 2)[1])
        ], key=itemgetter(0), reverse=True)

        flag = 0
        for opt in li[0][1].split(','):
            if opt in self.opts2statvfs_flag:
                flag = flag + self.opts2statvfs_flag[opt]
                st['f_flag'] = flag

        return st

    @full_path()
    def unlink(self, path, full_path):
        self._entry(self.cmd_builder.unlink(full_path))

    @full_path(['name', 'target'])
    def symlink(self, name, target, full_path_name, full_path_target):
        res = self._entry(self.cmd_builder.symlink(
            full_path_name,
            target
        ))
        return res.errno

    @full_path(['old', 'new'])
    def rename(self, old, new, full_path_old, full_path_new):
        res = self._entry(self.cmd_builder.rename(
            full_path_old, full_path_new))
        return res.errno

    @full_path(['name', 'target'])
    def link(self, name, target, full_path_name, full_path_target):
        res = self._entry(self.cmd_builder.link(
            full_path_name,
            full_path_target
        ))
        return res.errno

    @full_path()
    def utimens(self, path, times=None, full_path=None):
        if times is None:
            # set current time to full_path
            self._entry(self.cmd_builder.utimens_update_to_now(full_path))
        else:
            self._entry(
                self.cmd_builder.utimens_update_atime(
                    full_path, times[0]
                )
            )
            self._entry(
                self.cmd_builder.utimens_update_mtime(
                    full_path, times[1]
                )
            )

    def _read_dev_fd(self):
        """ only use to read `/dev/fd/`
        """
        res = self._entry(self.cmd_builder.readdir('/dev/fd/'))
        return re.sub(r'\n$', '', res.stdout).split('\n')

    @full_path()
    def open(self, path, flags, full_path):
        # Get available fd value by status of containers's '/dev/fd/'
        fd_set = set(fd for fd in self._read_dev_fd())
        fg = (i for i in range(20, 30) if not str(i) in fd_set)
        # TODO: Check no available fd then raise FuseOSError
        fd = next(fg)
        self._entry(self.cmd_builder.open(
                    fd, flags, shlex.quote(full_path)
                    ), quote=False, _disable_encode_stdoute_data=True)

        return fd

    @full_path()
    def create(self, path, mode, fi=None, full_path=None):
        self._entry(self.cmd_builder.create(full_path))
        self.chmod(path, mode)
        return self.open(path, os.O_WRONLY | os.O_CREAT)

    def read(self, path, length, offset, fh):
        # Always use fh
        res = self._entry(self.cmd_builder.read(
            length, offset, fh
        ), stdout_to_str=False)
        return res.stdout

    def write(self, path, buf, offset, fh):
        # Always use fh
        length = len(buf)
        self._entry(self.cmd_builder.write(
            length, offset, fh
        ), stdin_data=buf)
        return length

    @full_path()
    def truncate(self, path, length, fh=None, full_path=None):
        self._entry(self.cmd_builder.truncate(length, full_path))

    @full_path()
    def flush(self, path, fh, full_path=None):
        res = self._entry(
            self.cmd_builder.flush(fh, full_path),
            errno_map='fsync',
            raise_unknow=False
        )
        ret = res.errno if res.errno != 1 else 0
        return ret

    def release(self, path, fh):
        # Always use fh
        self._entry(self.cmd_builder.release(fh),
                    quote=False, _disable_encode_stdoute_data=True)
        return None

    def fsync(self, path, fdatasync, fh):
        return self.flush(path, fh)


if __name__ == '__main__':
    FUSE(XrosFS(sys.argv[1], sys.argv[2]),
         sys.argv[3], nothreads=True, foreground=True)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 hankei6km
# Licensed under the MIT License. See LICENSE.txt in the project root.

import sys
import logging
import logging.config
from .xrosfs import XrosFS
from docker import errors


def main(argv=sys.argv[1:]):
    import os
    from fuse import FUSE
    from .args import parse
    from .conshell import ConShell
    from .cmdbuilder import CmdBuilder

    mount_args = parse(argv)

    if mount_args.mount_opts.get('debug', False):
        # enable to output debug message from fuse.
        # (and from some modules. ie. docker etc.)
        logging.basicConfig(level=logging.DEBUG)

    shell = ConShell(mount_args.user)
    try:

        shell.connect(mount_args.container)
        cmd_builder = CmdBuilder()
        avail_cmds = shell.setup_avail_cmds(
            cmd_builder.require_cmds
        )
        cmd_builder.set_avail_cmds(avail_cmds)

        # normalize server_path(blank / rel path)
        server_path = mount_args.server_path
        if os.path.isabs(mount_args.server_path) is False:
            res = shell.entry(cmd_builder._get_homedir(), quote=False)
            server_path = os.path.join(
                res.stdout.split('\n', 1)[0],
                mount_args.server_path
            )
        server_path = os.path.normpath(server_path)

        # build fsname str(ie 'root@xros-bench:path/to')
        fsname = (mount_args.user if mount_args.user != '' else '') + \
            mount_args.container + ':' + mount_args.server_path

        FUSE(XrosFS(shell, cmd_builder, server_path),
             mount_args.mountpoint,
             nothreads=True,
             foreground=mount_args.foreground,
             # https://github.com/terencehonles/fusepy/wiki/Sharing-fusepy-mount-via-Samba
             # allow_other=True,
             fsname=fsname,
             **mount_args.mount_opts
             )

    except errors.NotFound as exc:
        print('No such container: {}'.format(mount_args.container))
        sys.exit(1)


__all__ = ['XrosFS', 'main']



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 hankei6km
# Licensed under the MIT License. See LICENSE.txt in the project root.

import os
import shlex
import base64

import docker

from .xros_utils import init_logger


class CmdSet:
    avail_cmd = {
        'which': {
            'path': '/bin/which'
        },
        'ls': {
            'path': '/bin/ls'
        },
        'echo': {
            'path': '/bin/echo'
        },
        'stat': {
            'path': '/usr/bin/stat'
        },
        'test': {
            'path': '/usr/bin/test'
        },
        'chmod': {
            'path': '/bin/chmod'
        },
        'chown': {
            'path': '/bin/chown'
        },
        'readlink': {
            'path': '/bin/readlink'
        },
        'rm': {
            'path': '/bin/rm'
        },
        'mkdir': {
            'path': '/usr/bin/mkdir'
        },
        'ln': {
            'path': '/usr/bin/ln'
        },
        'mv': {
            'path': '/bin/mv'
        },
        'exec': {
            'path': 'exec'
        },
        'touch': {
            'path': '/usr/bin/touch'
        },
        'dd': {
            'path': '/bin/dd'
        },
        'truncate': {
            'path': '/usr/bin/truncate'
        }
    }

    def get_fullpath(self, cmd_name):
        ret = ''
        try:
            ret = self.avail_cmd[cmd_name]['path']
        except KeyError:
            raise RuntimeError('not found cmd full path: {0}'.format(cmd_name))

        return ret


class EntryResult():
    __slots__ = ['stdout', 'errno', 'stderr']

    def __init__(self, stdout='', errno=0, stderr=''):
        self.stdout = stdout
        self.errno = errno
        self.stderr = stderr


class ConShell():
    """ Send shell command to container via `docker.exec` and
    Receive stdout/stderr from commands.
    """

    def __init__(self, user='', logger=init_logger(__name__)):
        self.user = user
        # config logger was proccessed in global(`_init__.py`).
        # It's followed fuse.LogginMixIn style.
        self.logger = logger
        # Suppress debug output method when debug mode is disabled.
        from logging import DEBUG
        if self.logger.isEnabledFor(DEBUG) is False:
            self._debug_entry_command = lambda command: None
            self._debug_entry_ret = lambda ret: None

    # _init_require_cmds = tuple('pwd')
    _avail_cmds = {
        # setup cmds that initialize other cmds in advance
        # (extend this dict at `connect`).
        'which': {
            'path': '/bin/which',
        },
        'exec': {
            'path': 'exec'  # function
        },
        'exit': {
            'path': 'exit'  # function
        }
    }
    _support_shell = ({
        'shell_name': 'bash',
        'shell_path': '/bin/bash',
        'shell_args': '--noprofile --norc',
        'prefix_command': 'LC_ALL=C set -o pipefail;'
    }, {
        'shell_name': 'ash',
        'shell_path': '/bin/ash',
        'shell_args': '--noprofile --norc',
        'prefix_command': 'LC_ALL=C set -o pipefail;'
    })

    _shell_name = 'sh'  # basic shell(replace other shell at `connect`)
    _shell_path = '/bin/sh'
    _shell_args = ''
    _prefix_command = 'LC_ALL=C '

    sh_stream = None

    def _debug_entry_command(self, command):
        self.logger.debug('#entry => {command}'.format(command=command))

    def _debug_entry_ret(self, ret):
        self.logger.debug('#entry <= errno={ret.errno}'.format(ret=ret))
        self.logger.debug('#entry <= stdout={ret.stdout}'.format(ret=ret))
        self.logger.debug('#entry <= stderr={ret.stderr}'.format(ret=ret))

    def _entry_command(self, command):
        # write data to readonly stream force.
        # (type(self.sh_stream) = <class 'socket.SocketIO'>)
        os.write(self.sh_stream.fileno(),
                 bytearray(command, 'utf-8'))

    def _read_result(self):
        """Read result from stream that connected container.
        Returns:
            fd: pipe's file descriptor(stdout, stderr)?
            result: data readed from stream.
        """
        # Header??
        # byte 0  : [0x01 or 0x02] pipe's file descriptor(stdout, stderr)?
        # byte 1-3: [0x00, 0x00, 0x00]
        # byte 4-7: Length?
        fd = self.sh_stream.read(1)
        self.sh_stream.read(3)
        body_len = int.from_bytes(self.sh_stream.read(4), byteorder='big')

        # Body
        result = self.sh_stream.read(body_len)

        return fd, result

    def _which_cmd(self, cmd_name):
        ret = ''

        # res = self.entry(('which', '--', cmd_name))
        for path in ['/usr/sbin', '/usr/bin', '/sbin', '/bin']:
            full_path = os.path.join(path, cmd_name)
            res = self.entry((
                'ls', full_path
            ))
            if res.errno == 0 and res.stdout.split('\n', 1)[0] == full_path:
                ret = full_path
                break

        return ret

    def _get_shell_item(self):
        ret = None

        for item in self._support_shell:
            res = self._which_cmd(item['shell_name'])
            if res != '':
                item['shell_path'] = res
                ret = item
                break

        return ret

    def _open_stream(self):
        self.sh_stream = self.stream_exec(
            ' '.join((self._shell_path, self._shell_args)))

    def _switch_to_support_shell(self):
        item = self._get_shell_item()
        if item is not None:
            # close current stream
            self.entry(['exit'])
            # setup shell attributes
            self._shell_name = item['shell_name']
            self._shell_path = item['shell_path']
            self._shell_args = item['shell_args']
            self._prefix_command = item['prefix_command']
            # open stream in new attributes
            self._open_stream()
        else:
            raise RuntimeError('not installed support shell'
                               'in target container')

    def _setup_init_shell(self):
        self._open_stream()

    def setup_avail_cmds(self, require_cmds):
        for cmd_name in require_cmds:
            cmd_path = self._which_cmd(cmd_name)
            if cmd_path != '':
                self._avail_cmds[cmd_name] = {  # consider define class??
                    'path': cmd_path
                }
            else:
                raise RuntimeError(
                    'not found cmd full path: {0}'.format(cmd_name)
                )
        return self._avail_cmds

    def connect(self, container_name):
        client = docker.from_env()
        try:
            self.container = client.containers.get(container_name)

            # setup initial shell to search bash/ash erc.
            # (i.e. run `$ test -x '/bin/bash`)
            self._setup_init_shell()

            # switch shell that 'sh' to support shell(bash/ash) and
            # re connect stream.
            self._switch_to_support_shell()
        except docker.errors.NotFound as exc:
            raise exc

        # print(self.sh_stream.writable()) # False
        # print(self.sh_stream.isatty()) # False
        # print(self.sh_stream.seekable()) # False
        # print(type(self.sh_stream)) # <class 'socket.SocketIO'>

    def quote(self, command):
        return [shlex.quote(i) for i in command]

    def entry(self, command, quote=True, stdin_data=None, stdout_to_str=True,
              mixin_stderr_to_stdout=False,
              _disable_encode_stdoute_data=False):
        ret = EntryResult()

        #
        # Prepare commands bytes
        #
        if quote:
            c = ' '.join(self.quote(command))
        else:
            c = ' '.join(command)
        self._debug_entry_command(command)

        prefix_command = self._prefix_command
        if stdin_data:
            base64_data = str(base64.b64encode(stdin_data), 'utf-8')
            prefix_command = prefix_command + 'echo -n '\
                + shlex.quote(base64_data)  \
                + '|base64 -d|'
        append_command = '\n'
        if not _disable_encode_stdoute_data:
            append_command = ' |base64' + append_command
        if mixin_stderr_to_stdout:
            append_command = ' 2>&1' + append_command

        #
        # Entry commands
        #

        self._entry_command(prefix_command + c + append_command)

        # if stdin_data:
        #     #self.sh_stream.flush()
        #     time.sleep(1)
        #     c = os.write(self.sh_stream.fileno(), stdin_data)
        #     #print('data:' + str(c))

        # Entry separetor.
        self._entry_command('/bin/echo -en "\\n${?}\\0\\0\\0:"\n')

        #
        # Retreve results
        #
        out_data = {
            b'\x01': b'',
            b'\x02': b''
        }

        fd, data = self._read_result()
        out_data[fd] = data
        chk = 'loop'
        while chk == 'loop':
            if out_data[b'\x01'][-4:] == b'\0\0\0:':
                # Trim delimiter bytes
                out_data[b'\x01'] = out_data[b'\x01'][0:-4]
                chk = 'done'
            else:
                fd, data = self._read_result()
                out_data[fd] = out_data[fd] + data

        # Split result from stdout to bytes and error code(as string).
        (stdout_result, errno_result) = out_data[b'\x01'].rsplit(b'\n', 1)
        ret.errno = int(str(errno_result, 'utf-8'))
        if not _disable_encode_stdoute_data:
            stdout_result = base64.b64decode(stdout_result)
        else:
            stdout_result = stdout_result
        if stdout_to_str:
            try:
                ret.stdout = str(stdout_result, 'utf-8')
            except UnicodeDecodeError:
                # capture exception to revocer?
                raise
        else:
            ret.stdout = stdout_result
        if len(out_data[b'\x02']) > 0:
            try:
                ret.stderr = str(out_data[b'\x02'], 'utf-8')
            except UnicodeDecodeError:
                # capture exception to revocer?
                raise
        self._debug_entry_ret(ret)
        return ret

    def stream_exec(self, command):
        """Exec command and reserve stream to container.

        Note:
            Always use to exec shell(sh) for prepare `exec` method.

        Args:
            command: command to docker exec.

        Returns:
            Stream to shell in container.

        """
        return self.container.exec_run(
            command,
            user=self.user,
            stdin=True,
            socket=True
        )


if __name__ == '__main__':  # pragma: no cover.
    import sys
    shell = ConShell()
    shell.connect(sys.argv[1])
    ret = shell.entry(['ls', 'test10'])
    print('stdout: ' + ret.stdout)
    print('errnor:' + str(ret.errno))
    print('stderr: ' + ret.stderr)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 hankei6km
# Licensed under the MIT License. See LICENSE.txt in the project root.


def init_logger(
        name,
        add_handler=False,
        timestamp=False,
        debug=False,
        verbose=False
):
    from logging import getLogger, \
        StreamHandler, \
        Formatter, \
        DEBUG, \
        INFO, \
        WARNING

    logger = getLogger(name)

    if add_handler:
        logger.propagate = False

        level = WARNING
        if debug:
            level = DEBUG
        elif verbose:
            level = INFO

        logger.setLevel(level=level)

        ch = StreamHandler()
        ch.setLevel(level)
        formater_str = '%(name)s:%(lineno)d: %(message)s'
        if timestamp:
            formater_str = '%(asctime)s %(name)s:%(lineno)d: %(message)s'
        formatter = Formatter(formater_str)
        ch.setFormatter(formatter)
        # ch.terminator = ''

        logger.addHandler(ch)

    return logger



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 hankei6km
# Licensed under the MIT License. See LICENSE.txt in the project root.

import sys
import os
import configparser
import pkg_resources
import argparse
import shlex
# from urllib.parse import urlparse
# import re

metadata = {
    'name': 'xrosfs',
    'description': 'Mount a Running Docker Container File Sytem via FUSE',
    'version': ''
}

try:
    script_path = os.path.dirname(os.path.realpath(__file__))
    config = configparser.ConfigParser()
    config.read(os.path.join(os.path.dirname(script_path), 'setup.cfg'))
    metadata['version'] = config['metadata']['version']
except KeyError:
    None

try:
    metadata['version'] = pkg_resources.get_distribution(
        metadata['name']
    ).version
except pkg_resources.DistributionNotFound:
    None

acceptable_mount_opts = ['allow_other', 'default_permissions', 'debug']


# https://stackoverflow.com/questions/41008269/python-argparse-assertionerror-from-metavar-userhostfile
class UsageFormatterTarget(argparse.HelpFormatter):
    def _format_usage(self, usage, actions, groups, prefix):
        result = super(UsageFormatterTarget, self)._format_usage(
            usage, actions, groups, prefix)
        return result.format(target='[user@]container:[dir]')


parser = argparse.ArgumentParser(prog=metadata['name'],
                                 description=metadata['description'],
                                 formatter_class=UsageFormatterTarget)

parser.add_argument('--version',
                    action='version',
                    version='%(prog)s version ' + metadata['version'])

parser.add_argument('-f',
                    dest='foreground',
                    action='store_true',
                    help='foreground operation')

parser.add_argument('-d',
                    dest='debug',
                    action='store_true',
                    help='enable debug output (implies -f)')

parser.add_argument('-o', '--opts',
                    dest='mount_opts',
                    metavar='opts',
                    action='append',
                    default=[],
                    help='mount options'
                    '(the only effective options is'
                    ' `{}` in present version)'.format(
                        ','.join(acceptable_mount_opts)
                    ))

parser.add_argument('target',
                    type=str,
                    # nargs=1,
                    # metavar='[user@]container:[dir]'
                    metavar='{target}'
                    )

parser.add_argument('mountpoint',
                    type=str)


class ArgsResult():
    __slots__ = [
        'foreground',
        'debug',
        'mount_opts',

        'user',
        'container',
        'server_path',
        'mountpoint'
    ]

    def __init__(self, foreground, debug, mount_opts,
                 user, container, server_path, mountpoint):
        self.foreground = foreground
        self.debug = debug
        self.mount_opts = mount_opts

        self.user = user
        self.container = container
        self.server_path = server_path
        self.mountpoint = mountpoint


def split_target(target):
    t = target.split('@', 1)
    if len(t) > 1:
        user = t[0]
        u = user.split(':', 1)
        if len(u) > 1:
            sys.exit('Password is not supported: ' + user)
        device = t[1]
    else:
        user = ''
        device = t[0]

    t = device.split(':')
    if len(t) == 2:
        container = t[0]
        server_path = t[1]
    else:
        sys.exit('Invalid format(bottom of container is not \':\'): ' + target)

    return (user, container, server_path)


def parse_opts(opts):
    ret = {}
    lexer = shlex.shlex(','.join(opts))
    lexer.whitespace = ','
    lexer.whitespace_split = True
    for o in lexer:
        kv = o.split('=', 1)
        k = kv[0]
        if k in acceptable_mount_opts:
            if len(kv) > 1:
                v = (kv[1])
                lv = v.lower()
                if lv == 'true':
                    v = True
                elif lv == 'false':
                    v = False
            else:
                v = True
            ret[k] = v
    return ret


def parse(argv):
    args = parser.parse_args(argv)

    # split target to "[user]" "container" "/path/to".
    user, container, server_path = split_target(args.target)

    # build opts dict(ie. `{'allow_other: True}`)
    mount_opts = parse_opts(args.mount_opts)

    # set debgu option by '-d'.
    if args.debug:
        mount_opts['debug'] = True

    return ArgsResult(
        foreground=args.foreground,
        debug=args.debug,
        mount_opts=mount_opts,
        user=user,
        container=container,
        server_path=server_path,
        mountpoint=args.mountpoint
    )


if __name__ == '__main__':  # pragma: no cover.
    mount_args = parse(sys.argv[1:])

    print('foreground: ' + str(mount_args.foreground))
    print('mount_opts: ' + str(mount_args.mount_opts))

    print('user: ' + mount_args.user)
    print('container: ' + mount_args.container)
    print('server_path: ' + mount_args.server_path)
    print('mountpoint: ' + mount_args.mountpoint)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 hankei6km
# Licensed under the MIT License. See LICENSE.txt in the project root.

import os
import datetime


class CmdBuilder:
    def __init__(self):
        self.stat2json = '{' + ','.join([
            '"st_mode":"0x%f"',  # output hexadecimal number as string.
            '"st_ino":%i',
            '"st_dev":%d',
            '"st_nlink":%h',
            '"st_uid":%u',
            '"st_gid":%g',
            '"st_size":%s',
            '"st_atime":%X',
            '"st_mtime":%Y',
            '"st_ctime":%Z',
        ]) + '}'

        self.statfs2json = '{' + ','.join([
            '"f_bavail":%a',
            '"f_bfree":%f',
            '"f_blocks":%b',
            '"f_bsize":%s',
            '"f_favail":%d',  # can't get to non-superuser in `stat`.
            '"f_ffree":%d',
            '"f_files":%c',
            # '"f_flag":%?', https://docs.python.org/3/library/os.html
            '"f_frsize":%S',
            '"f_namemax":%l',
        ]) + '}'

        self.chmod_mask = 4095
        # 4959 ==
        # stat.S_ISUID | stat.S_ISGID | stat.S_ENFMT | stat.S_ISVTX |
        # stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC | stat.S_IRWXU |
        # stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRWXG |
        # stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP | stat.S_IRWXO |
        # stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH

    require_cmds = (
        'pwd', 'ls', 'echo', 'stat', 'test', 'chmod', 'chown', 'readlink',
        'rm', 'rmdir', 'mkdir', 'ln', 'mv', 'touch', 'dd', 'truncate'
    )

    def set_avail_cmds(self, avail_cmds):
        self._avail_cmds = avail_cmds

    def _get_homedir(self):
        # This method does not use in fuse operation.
        # just use to get home dir of exec user
        # when not passed `server_path` from CLI args.
        return (
            self._avail_cmds['test']['path'],
            '-z',
            '"${HOME}"',
            '&&',
            self._avail_cmds['pwd']['path'],
            '||',
            self._avail_cmds['echo']['path'],
            '"${HOME}"'
        )

    def access(self, mode, full_path):
        return (self._avail_cmds['test']['path'], mode, full_path)

    def _chmod_octal_digit(self, mode):
        return '{0:04o}'.format(mode & self.chmod_mask)

    def chmod(self, mode, full_path=None):
        return (self._avail_cmds['chmod']['path'],
                self._chmod_octal_digit(mode),
                '--',
                full_path)

    def chown(self, uid, gid, full_path):
        arg = ''
        if uid != -1 and gid != -1:
            arg = str(uid) + ':' + str(gid)
        elif uid != -1 and gid == -1:
            arg = str(uid)
        elif uid == -1 and gid != -1:
            arg = ':' + str(gid)

        return (self._avail_cmds['chown']['path'], arg, '--', full_path)

    def getattr(self, full_path):
        return (
            self._avail_cmds['stat']['path'],
            '-c',
            self.stat2json,
            '--',
            full_path
        )

    def readdir(self, full_path):
        return (self._avail_cmds['ls']['path'], '-a1', '--', full_path)

    def readlink(self, full_path):
        return (self._avail_cmds['readlink']['path'], '-n', '--', full_path)

    def rmdir(self, full_path):
        # remove *empty* directories.
        return (self._avail_cmds['rmdir']['path'], '--', full_path)

    def mkdir(self, full_path):
        return (self._avail_cmds['mkdir']['path'], '--', full_path)

    def statfs(self, full_path):
        return (
            self._avail_cmds['stat']['path'],
            '-f', '-c', self.statfs2json, '--', full_path
        )

    def unlink(self, full_path):
        return (self._avail_cmds['rm']['path'], '--', full_path)

    def symlink(self, full_path_name, target):
        return (
            self._avail_cmds['ln']['path'],
            '-s',
            '--',
            target,
            full_path_name
        )

    def rename(self, full_path_old, full_path_new):
        return (
            self._avail_cmds['mv']['path'],
            '--',
            full_path_old,
            full_path_new
        )

    def link(self, full_path_name, full_path_target):
        return (
            self._avail_cmds['ln']['path'],
            '--',
            full_path_target, full_path_name
        )

    def utimens_update_to_now(self, full_path):
        return (
            self._avail_cmds['touch']['path'],
            '-c',
            '--',
            full_path
        )

    def utimens_update_atime(self, full_path, time):
        # TODO: detect acceptale args of touch in connected container and
        #       switch args to set timestamp.
        # '-a' flag not effect to touch of busybox.
        atime = datetime.datetime.fromtimestamp(time)
        tstr = atime.strftime('%Y%m%d%H%M.%S')
        return (self._avail_cmds['touch']['path'],
                '-c',
                '-a',
                '-t',
                tstr,
                '--',
                full_path)

    def utimens_update_mtime(self, full_path, time):
        # TODO: detect acceptale args of touch in connected container and
        #       switch args to set timestamp.
        # '-m' flag ignored to touch of busybox
        mtime = datetime.datetime.fromtimestamp(time)
        tstr = mtime.strftime('%Y%m%d%H%M.%S')
        return (self._avail_cmds['touch']['path'],
                '-c',
                '-m',
                '-t',
                tstr,
                '--',
                full_path)

    def open(self, fd, flags, full_path):
        flags = flags % 32768  # what mean is 32768(0x8000)?

        # TODO: Tranlate flags of os.open to operator strict.
        # It's not support `os.O_EXCL` etc yet.
        operator = '<'
        if os.O_RDONLY == flags:
            operator = '<'
        if os.O_WRONLY == flags:
            operator = '>'
        if os.O_CREAT == flags:
            operator = '>'
        if (os.O_WRONLY | os.O_TRUNC) == flags:
            operator = '>'
        if (os.O_WRONLY | os.O_APPEND) == flags:
            operator = '>>'
        if os.O_RDWR == flags:
            operator = '<>'
        if (os.O_RDWR | os.O_TRUNC) == flags:
            operator = '<>'
        if (os.O_RDWR | os.O_APPEND) == flags:
            operator = '<>>'

        return (
            self._avail_cmds['exec']['path'],
            str(fd) + operator,
            full_path
        )

    def create(self, full_path):
        return (self._avail_cmds['touch']['path'], '--', full_path)

    def read(self, length, offset, fh):

        return (self._avail_cmds['dd']['path'],
                'status=noxfer',  # busybox ver.
                # 'status=none',  # coreutil ver.
                'if=/dev/fd/' + str(fh),
                'bs=1',
                'count=' + str(length),
                'skip=' + str(offset))

    def write(self, length, offset, fh):
        return (self._avail_cmds['dd']['path'],
                'status=none',
                'of=/dev/fd/' + str(fh),
                'bs=1',
                'count=' + str(length),
                'seek=' + str(offset))

    def truncate(self, length, full_path):
        return (self._avail_cmds['truncate']['path'],
                '-s',
                str(length),
                '--',
                full_path)

    def flush(self, fh, full_path):
        # It is not confirmed whether `fsync` is actually called
        # return (['fsync', '--', full_path])
        return (self._avail_cmds['dd']['path'],
                'status=none',
                'of=' + full_path,
                'count=0',
                'conv=notrunc',
                'conv=fsync')

    def release(self, fh):
        return (self._avail_cmds['exec']['path'], str(fh) + '>&-')


if __name__ == '__main__':  # pragma: no cover.
    import shlex
    cmd_builder = CmdBuilder()
    c = ' '.join(
        [shlex.quote(i) for i in cmd_builder.getattr('/usr/bin/stat')])
    print(c)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 hankei6km
# Licensed under the MIT License. See LICENSE.txt in the project root.

from xrosfs import main

if __name__ == '__main__':
    import sys
    main(argv=sys.argv[1:])



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 hankei6km
# Licensed under the MIT License. See LICENSE.txt in the project root.

from setuptools import setup

# https://stackoverflow.com/questions/26737222/pypi-description-markdown-doesnt-work
try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()

setup(
    long_description=long_description,
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-asyncio', 'pytest-cov']
)
